@@ -17,7 +17,7 @@
 // 6.d input - attempt to read from sourcefile
 // 6.e feedback - calculate and display user feedback information
 // 6.f reaction - act according to result of read operation
-// 6.f.1 succesfull read:
+// 6.f.1 successful read:
 // 6.f.1.a attempt to backtrack for readable data prior to current position or...
 // 6.f.1.b write to output data file
 // 6.f.2 failed read
@@ -28,6 +28,8 @@
 
 #define _FILE_OFFSET_BITS 64
 // make off_t a 64 bit pointer on system that support it
+#define _GNU_SOURCE
+// allow usage of O_DIRECT if available
 
 #include <sys/types.h>
 #include <sys/wait.h>
@@ -36,6 +38,7 @@
 #include <errno.h>
 #include <fcntl.h>
 #include <stdio.h>
+#include <stdint.h>
 #include <unistd.h>
 #include <signal.h>
 #include <stdlib.h>
@@ -43,6 +46,14 @@
 #include <stdarg.h>
 #include "arglist.h"
 
+// define syncmode - use O_DIRECT if supported
+#ifdef O_DIRECT
+	#define O_SAFECOPYSYNC O_DIRECT
+#else
+	#define O_SAFECOPYSYNC O_RSYNC
+#endif
+
+
 #define DEBUG_FLOW 1
 #define DEBUG_IO 2
 #define DEBUG_BADBLOCKS 4
@@ -69,6 +80,7 @@
 #define DEFOUTPUTBB NULL
 #define DEFINPUTBB NULL
 #define DEFEXCLBB NULL
+#define DEFTIMINGFILESTRING NULL
 
 #define VERY_FAST 100
 #define FAST VERY_FAST*100
@@ -157,6 +169,8 @@ void usage(char * name) {
 	fprintf(stdout,"	                Linux   floppy        controller reset, twaddle\n");
 	fprintf(stdout,"	            Default: %i\n",DEFLOWLEVEL);
 	fprintf(stdout,"	--sync : Use synchronized read calls (disable driver buffering).\n");
+	fprintf(stdout,"	         Safecopy will use O_DIRECT if supported by the OS\n");
+	fprintf(stdout,"	         and O_SYNC otherwise.\n");
 	fprintf(stdout,"	         Default: Asynchronous read buffering by the OS is allowed\n");
 	fprintf(stdout,"	-s <blocks> : Start position where to start reading.\n");
 	fprintf(stdout,"	              Will correspond to position 0 in the destination file.\n");
@@ -207,6 +221,19 @@ void usage(char * name) {
 	fprintf(stdout,"	              Blocks not in the -I file, or covered by the file\n");
 	fprintf(stdout,"	              specified with -X are save from being overwritten.\n");
 	fprintf(stdout,"	              Default: none\n");
+	fprintf(stdout,"	--debug <level> : Enable debug output. Level is a bit field,\n");
+	fprintf(stdout,"	                  add values together for more information:\n");
+	fprintf(stdout,"	                     program flow:     %i\n",DEBUG_FLOW);
+	fprintf(stdout,"	                     IO control:       %i\n",DEBUG_IO);
+	fprintf(stdout,"	                     badblock marking: %i\n",DEBUG_BADBLOCKS);
+	fprintf(stdout,"	                     seeking:          %i\n",DEBUG_SEEK);
+	fprintf(stdout,"	                     incremental mode: %i\n",DEBUG_INCREMENTAL);
+	fprintf(stdout,"	                     exclude mode:     %i\n",DEBUG_EXCLUDE);
+	fprintf(stdout,"	                  or for all debug output: %i\n",255);
+	fprintf(stdout,"	                  Default: 0\n");
+	fprintf(stdout,"	-T <timingfile> : Write sector read timing information into\n");
+	fprintf(stdout,"	                  this file for later analysis.\n");
+	fprintf(stdout,"	                  Default: none\n");
 	fprintf(stdout,"	-h | --help : Show this text\n\n");
 	fprintf(stdout,"Valid parameters for -f -r -b <size> options are:\n");
 	fprintf(stdout,"	<integer>	Amount in bytes - i.e. 1024\n");
@@ -295,6 +322,7 @@ long int timediff(struct timeval oldtime,struct timeval newtime) {
 
 	long int usecs=newtime.tv_usec-oldtime.tv_usec;
 	usecs=usecs+((newtime.tv_sec-oldtime.tv_sec)*1000000);
+	if (usecs<0) usecs=0;
 	return usecs;
 
 }
@@ -408,14 +436,14 @@ void markbadblocks(int destination, off_t writeposition, off_t remain, char* mar
 		cposition=lseek(destination,writeposition,SEEK_SET);
 		if (cposition<0) {
 			fprintf(stderr,"\nError: seek() in output failed");
-			perror("");
+			perror(" ");
 			return;
 		}
 		debug(DEBUG_IO,"debug: writing badblock marker to destination file: %llu bytes at %llu\n",writeremain,cposition);
 		writeblock=write(destination,databuffer+(writeoffset % blocksize),(blocksize>writeremain?writeremain:blocksize));
 		if (writeblock<=0) {
 			fprintf(stderr,"\nError: write to output failed");
-			perror("");
+			perror(" ");
 			return;
 		}
 		writeremain-=writeblock;
@@ -477,7 +505,7 @@ void realmarkoutput (
 				*xblocksin=fopen(xblocksinfile,"r");
 				if (*xblocksin==NULL) {
 					fprintf(stderr,"Error reopening exclusion badblock file for reading: %s",xblocksinfile);
-					perror("");
+					perror(" ");
 					*previousxblock=((unsigned)-1)>>1;
 					*lastxblock=((unsigned)-1)>>1;
 					return;
@@ -613,13 +641,14 @@ int main(int argc, char ** argv) {
 	char *bblocksoutstring=DEFOUTPUTBB;
 	char *xblocksinstring=DEFEXCLBB;
 	char *failuredefstring=DEFFAILSTRING;
+	char *timingfilestring=DEFTIMINGFILESTRING; 
 	int retriesdef=DEFRETRIES;
 	int headmovedef=DEFHEADMOVE;
 	int lowleveldef=DEFLOWLEVEL;
 	// file descriptors
 	int source,destination,bblocksout;
 	// high level file descriptor
-	FILE *bblocksin,*xblocksin;
+	FILE *bblocksin,*xblocksin,*timingfile;
 
 	// file offset variables
 	off_t readposition,cposition,sposition,writeposition;
@@ -627,7 +656,7 @@ int main(int argc, char ** argv) {
 	// variables for handling read/written sizes/remainders
 	off_t remain,maxremain,block,writeblock,writeremain;
 	// pointer to main IO data buffer
-	char * databuffer;
+	char * databuffer,*databufferpool;
 	// a buffer for output text
 	char textbuffer[256];
 	// buffer pointer for sfgets() 
@@ -690,6 +719,7 @@ int main(int argc, char ** argv) {
 	arglist_addarg (carglist,"-S",1);
 	arglist_addarg (carglist,"-Z",1);
 	arglist_addarg (carglist,"-M",1);
+	arglist_addarg (carglist,"-T",1);
 
 	// find out wether the user is wetware
 	human=(isatty(1) & isatty(2));
@@ -748,8 +778,12 @@ int main(int argc, char ** argv) {
 	// synchronous IO
 	syncmode=0;
 	if (arglist_arggiven(carglist,"--sync")==0) {
-		fprintf(stdout,"Using synchronized IO on source.\n");
-		syncmode=O_RSYNC;
+		if (O_SAFECOPYSYNC==O_SYNC) {
+			fprintf(stdout,"Using synchronized IO on source. (O_SYNC)\n");
+		} else {
+			fprintf(stdout,"Using synchronized IO on source. (O_DIRECT)\n");
+		}
+		syncmode=O_SAFECOPYSYNC;
 	}
 
 	// find out source file size and block size
@@ -772,7 +806,7 @@ int main(int argc, char ** argv) {
 		fprintf(stdout,"File size: %llu\n",filesize);
 	} else {
 		fprintf(stderr,"Filesize not reported by stat(), trying seek().\n");
-		source=open(sourcefile,O_RDONLY | syncmode);
+		source=open(sourcefile,O_RDONLY | O_RSYNC);
 		if (source) {
 			filesize=lseek(source,0,SEEK_END);
 			close(source);
@@ -881,6 +915,11 @@ int main(int argc, char ** argv) {
 		fprintf(stdout,"Seek script (fallback): %s\n",seekscriptfile);
 	}
 
+	if (arglist_arggiven(carglist,"-T")==0) {
+		timingfilestring=arglist_parameter(carglist,"-T",0);
+		fprintf(stdout,"Write sector timing information to file: %s\n",timingfilestring);
+	}
+
 	if (arglist_arggiven(carglist,"-M")==0) {
 		failuredefstring=arglist_parameter(carglist,"-M",0);
 		if (failuredefstring==NULL) failuredefstring="";
@@ -911,11 +950,13 @@ int main(int argc, char ** argv) {
 		filesize=startoffset+length;
 	}
 
-	databuffer=(char*)malloc((blocksize+1)*sizeof(char));
-	if (databuffer==NULL) {
-		perror("MEMORY ALLOCATION ERROR!\nCOULDNT ALLOCATE MAIN BUFFER");
+	databufferpool=(char*)malloc(((2*blocksize)+1)*sizeof(char));
+	if (databufferpool==NULL) {
+		perror("MEMORY ALLOCATION ERROR!\nCOULDNT ALLOCATE MAIN BUFFER ");
 		return 2;
 	}
+	// align databuffer to block size - needed to support O_DIRECT and /dev/raw devices
+	databuffer=databufferpool+blocksize-(((uintptr_t)databufferpool)%blocksize);
 
 // 3.opening io handles
 		
@@ -924,7 +965,7 @@ int main(int argc, char ** argv) {
 	source=open(sourcefile,O_RDONLY | O_NONBLOCK | syncmode );
 	if (source==-1) {
 		fprintf(stderr,"Error opening sourcefile: %s",sourcefile);
-		perror("");
+		perror(" ");
 		if (human) usage(argv[0]);
 		arglist_kill(carglist);
 		return 2;
@@ -934,7 +975,7 @@ int main(int argc, char ** argv) {
 		if (xblocksin==NULL) {
 			close(source);
 			fprintf(stderr,"Error opening exclusion badblock file for reading: %s",xblocksinfile);
-			perror("");
+			perror(" ");
 			arglist_kill(carglist);
 			return 2;
 		}
@@ -946,7 +987,7 @@ int main(int argc, char ** argv) {
 			close(source);
 			if (excluding==1) fclose(xblocksin);
 			fprintf(stderr,"Error opening badblock file for reading: %s",bblocksinfile);
-			perror("");
+			perror(" ");
 			arglist_kill(carglist);
 			return 2;
 		}
@@ -956,7 +997,7 @@ int main(int argc, char ** argv) {
 			fclose(bblocksin);
 			if (excluding==1) fclose(xblocksin);
 			fprintf(stderr,"Error opening destination: %s",destfile);
-			perror("");
+			perror(" ");
 			if (human) usage(argv[0]);
 			arglist_kill(carglist);
 			return 2;
@@ -981,7 +1022,7 @@ int main(int argc, char ** argv) {
 			close(source);
 			if (excluding==1) fclose(xblocksin);
 			fprintf(stderr,"Error opening destination: %s",destfile);
-			perror("");
+			perror(" ");
 			if (human) usage(argv[0]);
 			arglist_kill(carglist);
 			return 2;
@@ -995,7 +1036,23 @@ int main(int argc, char ** argv) {
 			if (incremental==1) fclose(bblocksin);
 			if (excluding==1) fclose(xblocksin);
 			fprintf(stderr,"Error opening badblock file for writing: %s",bblocksoutfile);
-			perror("");
+			perror(" ");
+			arglist_kill(carglist);
+			return 2;
+		}
+	}
+
+	timingfile=NULL;
+	if (timingfilestring!=NULL) {
+		timingfile=fopen(timingfilestring,"w");
+		if (timingfile==NULL) {
+			close(source);
+			close(destination);
+			if (incremental==1) fclose(bblocksin);
+			if (excluding==1) fclose(xblocksin);
+			if (bblocksout) close(bblocksout);
+			fprintf(stderr,"Error opening timingfile for writing: %s",bblocksoutfile);
+			perror(" ");
 			arglist_kill(carglist);
 			return 2;
 		}
@@ -1019,7 +1076,7 @@ int main(int argc, char ** argv) {
 		//we are currently dealing with an unrecoverable error and need to find the end of the bad area
 	newsofterror=0; //flag that indicates previous read failures on the current block
 	lasterror=0; //address of the last encountered bad area in source file
-	lastgood=0; //address of the last encountered succesfull read in source file
+	lastgood=0; //address of the last encountered successful read in source file
 	lastmarked=0; //last known address already marked in output file
 	lastbadblock=-1; //most recently encountered block for output badblock file
 	lastxblock=-1; //most recently encountered block number from input exclude file
@@ -1044,7 +1101,7 @@ int main(int argc, char ** argv) {
 	// attempt to seek to start position to find out wether source file is seekable
 	cposition=lseek(source,startoffset,SEEK_SET);
 	if (cposition<0) {
-		perror("Warning: Input file is not seekable");
+		perror("Warning: Input file is not seekable ");
 		seekable=0;
 		close(source);
 		cposition=emergency_seek(startoffset,0,blocksize,seekscriptfile);
@@ -1053,11 +1110,12 @@ int main(int argc, char ** argv) {
 		}
 		source=open(sourcefile,O_RDONLY | O_NONBLOCK | syncmode);
 		if (source==-1) {
-			perror("Error reopening sourcefile after external seek");
+			perror("Error reopening sourcefile after external seek ");
 			close(destination);
 			if (incremental==1) fclose(bblocksin);
 			if (excluding==1) fclose(xblocksin);
 			if (bblocksoutfile!=NULL) close(bblocksout);
+			if (timingfile!=NULL) fclose(timingfile);
 			arglist_kill(carglist);
 			return 2;
 		}
@@ -1140,7 +1198,7 @@ int main(int argc, char ** argv) {
 					if (xblocksin==NULL) {
 						excluding=0;
 						fprintf(stderr,"Error reopening exclusion badblock file for reading: %s",xblocksinfile);
-						perror("");
+						perror(" ");
 						wantabort=1;
 						break;
 					}
@@ -1246,11 +1304,12 @@ int main(int argc, char ** argv) {
 						// reopen input file
 						source=open(sourcefile,O_RDONLY | O_NONBLOCK | syncmode );
 						if (source==-1) {
-							perror("\nError reopening sourcefile after external seek");
+							perror("\nError reopening sourcefile after external seek ");
 							close(destination);
 							if (incremental==1) fclose(bblocksin);
 							if (excluding==1) fclose(xblocksin);
 							if (bblocksoutfile!=NULL) close(bblocksout);
+							if (timingfile!=NULL) fclose(timingfile);
 							arglist_kill(carglist);
 							return 2;
 						}
@@ -1319,6 +1378,9 @@ int main(int argc, char ** argv) {
 		// time reading for quality calculation
 		gettimeofday(&newtime,NULL);
 		elapsed=timediff(oldtime,newtime);
+		if (timingfile) {
+			fprintf(timingfile,"%010llu %lu %lli\n",sposition/blocksize,elapsed,block);
+		}
 
 // 6.e feedback - calculate and display user feedback information
 
@@ -1360,17 +1422,17 @@ int main(int argc, char ** argv) {
 		output=0;
 
 // 6.f processing - react according to result of read operation
-// 6.f.1 succesfull read:
+// 6.f.1 successful read:
 		if (block>0) {
-			debug(DEBUG_IO,"debug: succesfull read\n");
+			debug(DEBUG_IO,"debug: successful read\n");
 			sposition=sposition+block;
-			// successfull read, if happening during soft recovery
+			// successful read, if happening during soft recovery
 			// (downscale or retry) list a single soft error
 			if (newsofterror==1) {
 				newsofterror=0;
 				softerr++;
 			}
-			// read successfull, test for end of damaged area
+			// read successful, test for end of damaged area
 			if (newerror==0) {
 // 6.f.1.a attempt to backtrack for readable data prior to current position or...
 				// we are in recovery since we just read past the end of a damaged area
@@ -1439,12 +1501,13 @@ int main(int argc, char ** argv) {
 					cposition=lseek(destination,writeposition,SEEK_SET);
 					if (cposition<0) {
 						fprintf(stderr,"\nError: seek() in %s failed",destfile);
-						perror("");
+						perror(" ");
 						close(destination);
 						close(source);
 						if (incremental==1) fclose(bblocksin);
 						if (excluding==1) fclose(xblocksin);
 						if (bblocksoutfile!=NULL) close(bblocksout);
+						if (timingfile!=NULL) fclose(timingfile);
 						arglist_kill(carglist);
 						return 2;
 					}
@@ -1452,12 +1515,13 @@ int main(int argc, char ** argv) {
 					writeblock=write(destination,databuffer+writeoffset,writeremain);
 					if (writeblock<=0) {
 						fprintf(stderr,"\nError: write to %s failed",destfile);
-						perror("");
+						perror(" ");
 						close(destination);
 						close(source);
 						if (incremental==1) fclose(bblocksin);
 						if (excluding==1) fclose(xblocksin);
 						if (bblocksoutfile!=NULL) close(bblocksout);
+						if (timingfile!=NULL) fclose(timingfile);
 						arglist_kill(carglist);
 						return 2;
 					}
@@ -1468,7 +1532,7 @@ int main(int argc, char ** argv) {
 			}
 		} else if (block<0) {
 // 6.f.2 failed read
-			debug(DEBUG_IO,"debug: read failed\n");
+			debug(DEBUG_IO,"debug: read failed: %s\n",strerror(errno));
 			// operation failed
 			counter=1;
 			// use low level IO for error correction if allowed
@@ -1542,27 +1606,29 @@ int main(int argc, char ** argv) {
 			// do some forced seeks to move head around.
 			for (cseeks=0;cseeks<seeks;cseeks++) {
 				debug(DEBUG_SEEK,"debug: forced head realignment\n");
+				// note. must use O_RSYNC since with O_DIRECT / raw devices, lseek to end of file might not work
 				source=open(sourcefile,O_RDONLY|O_RSYNC );
 				if (source) {
 					lseek(source,0,SEEK_SET);
-					read(source,&textbuffer,1);
+					read(source,databuffer,blocksize);
 					close(source);
 				}
 				source=open(sourcefile,O_RDONLY|O_RSYNC );
 				if (source) {
 					lseek(source,-blocksize,SEEK_END);
-					read(source,&textbuffer,1);
+					read(source,databuffer,blocksize);
 					close(source);
 				}
 				if (wantabort) break;
 			}
 			source=open(sourcefile,O_RDONLY | O_NONBLOCK | syncmode );
 			if (source==-1) {
-				perror("\nError reopening sourcefile after read error");
+				perror("\nError reopening sourcefile after read error ");
 				close(destination);
 				if (incremental==1) fclose(bblocksin);
 				if (excluding==1) fclose(xblocksin);
 				if (bblocksoutfile!=NULL) close(bblocksout);
+				if (timingfile!=NULL) fclose(timingfile);
 				arglist_kill(carglist);
 				return 2;
 			}
@@ -1580,7 +1646,7 @@ int main(int argc, char ** argv) {
 	fflush(stdout);
 	fflush(stderr);
 	if (newerror==0) {
-		// if theres an error at the end of input, treat as if we had one succesfull read afterwards
+		// if theres an error at the end of input, treat as if we had one successful read afterwards
 		tmp_pos=readposition/blocksize;
 		tmp_bytes=readposition-lastgood;
 		damagesize+=tmp_bytes;
@@ -1608,6 +1674,7 @@ int main(int argc, char ** argv) {
 	if (incremental==1) fclose(bblocksin);
 	if (excluding==1) fclose(xblocksin);
 	if (bblocksoutfile!=NULL) close(bblocksout);
+	if (timingfile!=NULL) fclose(timingfile);
 	arglist_kill(carglist);
 	return(0);
 }
